import os
import time
import ray
import traceback
from loguru import logger
from .cluster_status_manager import ClusterStatusManager
from .utils import get_self_addr, run_shell_command, ray_nodes_info_formatter

class ClusterLauncher:

    def __init__(
            self,
            cluster_nodes_count: int,
            head_node_addr: str,
            head_node_ray_port: int = 54321,
            head_node_ray_dashboard_port: int = 54322,
            node_rank: int = None,
            node_ray_manager_port: int = 54323,
            validate_head_node_addr: bool = True,
            export_env_var_names: list = None,
            head_node_spin_wait_interval: int = 3,
            worker_node_spin_wait_interval: int = 3,
    ) -> None:
        
        
        assert isinstance(cluster_nodes_count, int)
        self.cluster_nodes_count = cluster_nodes_count
        self.head_node_addr = head_node_addr
        self.head_node_ray_port = head_node_ray_port
        self.head_node_ray_dashboard_port = head_node_ray_dashboard_port

        # get node rank from arg or environ
        if node_rank is None:
            node_rank_in_env_var = os.environ.get("NODE_RANK")
            if node_rank_in_env_var is not None:
                self.node_rank = int(node_rank_in_env_var)
            else:
                raise ValueError("must specify node rank either by passing to args or exporting to environ")
        else:
            assert isinstance(node_rank, int)
            assert node_rank < cluster_nodes_count and node_rank >= 0
            self.node_rank = node_rank
        
        self.node_ray_manager_port = node_ray_manager_port

        # get node addr and optionally validate it
        self.node_addr = get_self_addr()
        if self.node_rank == 0:
            if validate_head_node_addr:
                assert self.head_node_addr == self.node_addr, (
                    "head node has rank = 0, this is rank 0 node but got"
                    f"ip {self.node_addr}, expect {self.head_node_addr}"
                )
            self.is_head_node = True
        else:
            if validate_head_node_addr:
                assert self.head_node_addr != self.node_addr, (
                    f"head node has rank = 0, this is rank {self.node_rank} node "
                    f"but got head node addr of {self.head_node_addr}"
                )
            self.is_head_node = False
        
        self.cluster_status_manager_handle = None
        self.cluster_status_manager_actor_name = "ClusterStatusManager"
        self.cluster_namespace = "ClusterLauncher"

        self.export_env_var_names = ["PYTHONPATH", "CUDA_VISIBLE_DEVICES"]
        if export_env_var_names is not None:
            self.export_env_var_names.extend(export_env_var_names)
        
        self.head_node_spin_wait_interval = head_node_spin_wait_interval
        self.worker_node_spin_wait_interval = worker_node_spin_wait_interval
            
    
    # context protocol
    def __enter__(self):
        node_started = self._start_ray_node()

        # step1: head node waiting for workers to join the cluster
        if self.is_head_node:
            logger.debug(f"setting up cluster status manager")
            self.cluster_status_manager_handle = \
                ClusterStatusManager.options(
                    name=self.cluster_status_manager_actor_name
                ).remote(self.cluster_nodes_count)
            ray.get(
                self.cluster_status_manager_handle.set_node_validity.remote(self.node_rank, node_started)
            )
        else:
            while True:
                try:
                    self.cluster_status_manager_handle = ray.get_actor(self.cluster_status_manager_actor_name)
                    break
                except Exception as e:
                    logger.debug(f"waiting for cluster status manager ({e})")
                    time.sleep(self.worker_node_spin_wait_interval)
            
            ray.get(self.cluster_status_manager_handle.set_node_validity.remote(self.node_rank, node_started))
        
        # step2: return context on head node, worker nodes spin wait cluster to tear down
        while True:
            
            if self.is_head_node:
                if ray.get(self.cluster_status_manager_handle.is_cluster_ready.remote()):
                    # once the cluster is ready, return self on head node
                    return self
                else:
                    # if is not ready, spin wait
                    time.sleep(self.head_node_spin_wait_interval)
            else:
                if ray.get(self.cluster_status_manager_handle.is_cluster_tearing_down.remote()):
                    # if it is tearing down, stop self ans exit
                    ray.get(self.cluster_status_manager_handle.set_node_validity.remote(self.node_rank, False))
                    self._stop_ray_node()
                    exit(0)
                else:
                    time.sleep(self.worker_node_spin_wait_interval)        

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if exc_type is not None:
            s = f"exiting context with exception, exception type: {exc_type}, exception value: {exc_value}, "
            formatted_traceback = ''.join(traceback.format_exception(exc_type, exc_value, exc_traceback))
            s += f"exception traceback: {formatted_traceback}"
            logger.error(s)
        else:
            logger.debug("exiting context with no exception")
        
        if self.is_head_node:
            ray.get(self.cluster_status_manager_handle.set_node_validity.remote(self.node_rank, False))
            while True:
                try:
                    if ray.get(self.cluster_status_manager_handle.is_cluster_finished.remote()):
                        break
                    else:
                        logger.debug(f"tearing down the cluster, waiting for worker nodes to exit")
                        time.sleep(self.head_node_spin_wait_interval)
                except:
                    break
            self._stop_ray_node()
            logger.info(f"cluster of {self.cluster_nodes_count} nodes stopped successfully")
        else:
            # only head node will reach here
            raise RuntimeError("worker nodes should not reach here")


    def _start_ray_node(self):
        # step1: run ray start command
        logger.debug(f"runnnig ray start command on node {self.node_rank}")
        if self.is_head_node:
            command = (
                "RAY_prestart_worker_first_driver=0 "
                "ray start --head --include-dashboard=True "
                f"--node-ip-address={self.head_node_addr} --port={self.head_node_ray_port} "
                f"--node-name='ray-head-node-rank-{self.node_rank}-{self.node_addr}' "
                f"--dashboard-port={self.head_node_ray_dashboard_port} --node-manager-port {self.node_ray_manager_port}"
            )
        else:
            command = (f"ray start --address={self.head_node_addr}:{self.head_node_ray_port} "
                    f"--node-name='ray-worker-node-rank-{self.node_rank}-{self.node_addr}' "
                    f"--node-manager-port {self.node_ray_manager_port} "
            )
        ret_code, *_ = run_shell_command(command)
        if ret_code != 0:
            return False
        
        # step2: run ray.init
        env_vars = {}
        for name in self.export_env_var_names:
            if name in os.environ:
                env_vars.update(
                    {name: os.environ.get(name)}
                )
            else:
                logger.warning(f"{name} does not exist in environ")
        logger.debug(f"running ray init on node {self.node_rank} with env: {env_vars}")
        try:
            ray.init(
                runtime_env={"env_vars": env_vars}, 
                namespace=self.cluster_namespace, 
                _node_ip_address=self.head_node_addr, 
                log_to_driver=True
            )
            # must set namespace, or the actor cannot get by name
            logger.debug(f"ray init on node {self.node_rank} finished")
        except Exception as e:
            logger.error(f"ray init on node {self.node_rank} failed due to {e}")
            return False
        
        # step3: head node spin wait for all nodes
        if self.is_head_node:
            alive_nodes = 0
            while alive_nodes < self.cluster_nodes_count:
                nodes = ray.nodes()
                alive_nodes = sum(1 for node in nodes if node["Alive"] == True)
                logger.debug(f"waiting for cluster nodes to start, {alive_nodes}/{self.cluster_nodes_count} ray nodes started")
                time.sleep(self.head_node_spin_wait_interval)
        # _ = run_shell_command("ray status")
        logger.info(ray_nodes_info_formatter(ray.nodes()))

        return True
    
    def _stop_ray_node(self):
        try:
            ray.shutdown()
            logger.debug(f"shutting down ray on node {self.node_rank}")
        except Exception as e:
            logger.debug(f"shutting down ray on node {self.node_rank} failed due to {e}")
        ret_code, *_ = run_shell_command("ray stop")
        if ret_code == 0:
            logger.debug(f"ray stopped on node {self.node_rank}")



import ray
import inspect

from collections import namedtuple
from functools import partial
from loguru import logger
from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy

from .base_backend import BaseBackend

PlacementGroupAndIndex = namedtuple("PlacementGroupAndIndex", ["placement_group", "bundle_index"])

class RemoteModule:
    def __init__(
            self, 
            backend_clz,
            placement_groups_and_indices: list[PlacementGroupAndIndex],
            discrete_gpu_actors: bool, # must be gpu actor first, cpu actor is not discrete
            module_name: str = None
    ):
        self.backend_clz = backend_clz
        assert issubclass(self.backend_clz, BaseBackend)
        if module_name is None:
            module_name = backend_clz.__name__ + str(id(self))
        self.module_name = module_name

        self.discrete_gpu_actors = discrete_gpu_actors
        
        self.backend_actors = []
        self._create_backend_actors(placement_groups_and_indices)

        self._register_remote_funcs()

    
    def _create_backend_actors(self, placement_groups_and_indices: list[PlacementGroupAndIndex]):
        if self.discrete_gpu_actors is True:
            for pg, idx in placement_groups_and_indices:
                current_bundle_gpu_count = int(pg.bundle_specs[idx].get("GPU"))
                assert current_bundle_gpu_count > 0, f"discrete gpu actor must be created on group with gpu resource"
                current_bundle_cpu_count_per_gpu = float(pg.bundle_specs[idx].get("CPU"))/current_bundle_gpu_count
                for _ in range(current_bundle_gpu_count):
                    remote_actor = ray.remote(
                            num_gpus=1,
                            num_cpus=current_bundle_cpu_count_per_gpu
                        )(self.backend_clz).options(
                            scheduling_strategy=PlacementGroupSchedulingStrategy(
                                placement_group=pg,
                                placement_group_bundle_index=idx,
                        )# , runtime_env={"env_vars": {"RAY_DEDUP_LOGS": "0"}}
                        ).remote()
                    self.backend_actors.append(remote_actor)
                    logger.debug(f"created remote actor {len(self.backend_actors) - 1} of module {self.module_name} on {pg.id} idx={idx} with 1 gpu and {current_bundle_cpu_count_per_gpu} cpu")
            assert len(self.backend_actors) > 0
            rank_0_actor = self.backend_actors[0]
            module_master_addr = ray.get(rank_0_actor.get_ip_address.remote())
            module_master_port = ray.get(rank_0_actor.get_avaiable_port.remote())
            logger.debug(f"rank 0 backend gives {module_master_addr=}, {module_master_port=}")

            set_environs_futures = []
            for actor_idx, actor in enumerate(self.backend_actors):
                set_environs_futures.append(actor.set_distributed_environs.remote(
                    actor_idx,
                    len(self.backend_actors),
                    module_master_addr,
                    module_master_port
                ))
            ray.get(set_environs_futures)


        else:
            assert len(placement_groups_and_indices) == 1, f"the actor is continuous, should not spread to groups"
            pg, idx = placement_groups_and_indices.pop()
            current_bundle_gpu_count = int(pg.bundle_specs[idx].get("GPU"))
            current_bundle_cpu_count = int(pg.bundle_specs[idx].get("CPU"))
            self.backend_actors.append(
                ray.remote(
                    num_gpus=current_bundle_gpu_count,
                    num_cpus=current_bundle_cpu_count
                )(self.backend_clz).options(
                    scheduling_strategy=PlacementGroupSchedulingStrategy(
                        placement_group=pg,
                        placement_group_bundle_index=idx,
                ) # , runtime_env={"env_vars": {"RAY_EXPERIMENTAL_NOSET_CUDA_VISIBLE_DEVICES": "1",}}
                ).remote()
            )
            logger.debug(f"created single remote actor of module {self.module_name} on {pg.id} idx={idx} with {current_bundle_gpu_count} gpu and {current_bundle_cpu_count} cpu")
    

    def _call_func_of_all_remote_actors(self, func_name: str, *args, **kwargs):
        all_func_returns = []
        for actor in self.backend_actors:
            assert hasattr(actor, func_name)
            all_func_returns.append(getattr(actor, func_name).remote(*args, **kwargs))
        if len(all_func_returns) == 1:
            all_func_returns = all_func_returns[0]
        else:
            logger.debug(f"module {self.module_name} contains multiple actors, will return a list of all results")
        return ray.get(all_func_returns)
    
    
    def _register_remote_funcs(self):
        self.remote_funcs = []
        for name, member in inspect.getmembers(self.backend_clz, predicate=inspect.isfunction):
            if not name.startswith("__"): # auto register all non-magic methods
                self.remote_funcs.append(name)
                setattr(self, name, partial(self._call_func_of_all_remote_actors, name))
                logger.debug(f"auto detected and registered remote func: {name}({member})")




import socket
import subprocess
from loguru import logger

def get_self_addr():
    try:
        hostname = socket.gethostname()
        local_ip = socket.gethostbyname(hostname)
        return local_ip
    except Exception as e:
        logger.warning(f"failed to get addr by hostname, due to: {e}")
    
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(("8.8.8.8", 80))
        ip = s.getsockname()[0]
        s.close()
        return ip
    except Exception as e:
        logger.warning(f"failed to get addr by connect to external, due to {e}")
    finally:
        s.close()
    
    logger.error(f"cannot get current node's addr")
    raise RuntimeError

def run_shell_command(cmd):
    logger.debug(f"executeing command: >>>{cmd}")
    retval = subprocess.run(cmd, shell=True, capture_output=True, text=True, check=False)
    if retval.returncode == 0:
        logger.debug(retval.stdout)
    else:
        logger.error(f"command returns non-zero value: {retval.stderr}")
    return  retval.returncode, retval.stdout, retval.stderr




def ray_nodes_info_formatter(ray_nodes):
    
    def _ray_node_formatter(node):
        is_alive = node.get('Alive', False)
        s = f"Node {node.get('NodeID', 'N/A')} ({node.get('NodeName', 'N/A')}) [{'ALIVE' if is_alive else 'DEAD'}]\n"
        s += f"NodeManager: {node.get('NodeManagerAddress', 'N/A')}:{node.get('NodeManagerPort', 'N/A')} ({node.get('NodeManagerHostname', 'N/A')})\n"
        s += f"ObjectStoreSocketName: {node.get('ObjectStoreSocketName', 'N/A')}, ObjectManagerPort: {node.get('ObjectManagerPort', 'N/A')}\n"
        s += f"RayletSocketName: {node.get('RayletSocketName', 'N/A')}\n"
        if is_alive:
            s += f"Resources:\n"
            for resource, value in node.get('Resources', {}).items():
                s += f"  {resource}: {value}\n"
        else:
            s += f"DeathReason: {node.get('DeathReason', 'N/A')}\n"
            s += f"DeathReasonMessage: {node.get('DeathReasonMessage', '')}\n"
        
        s += "\n"
        return s
    
    s = f"ray nodes information: \n"
    for node in ray_nodes:
        s += _ray_node_formatter(node)
    
    return s


import ray
from enum import Enum
from loguru import logger

class ClusterStatus(Enum):
    INIT = 0
    READY = 1
    TEARING_DOWN = 2
    FINISHED = 3

@ray.remote
class ClusterStatusManager:
    def __init__(self, cluster_nodes_count):
        self.nodes_validity = {}
        self.cluster_nodes_count = cluster_nodes_count
        self.cluster_status = ClusterStatus.INIT
    
    def count_valid_nodes(self):
        return sum(self.nodes_validity.values())
    
    def set_node_validity(self, node_rank, validity):
        if node_rank not in self.nodes_validity:
            self.nodes_validity[node_rank] = False
        self.nodes_validity[node_rank] = validity
        logger.debug(f"setting node {node_rank} validity as {validity}")

        self._cluster_status_update()
    

    def _cluster_status_update(self):
        valid_node_count = self.count_valid_nodes()
        if self.cluster_status == ClusterStatus.INIT:
            logger.debug(f"valid nodes in cluster: {valid_node_count}/{self.cluster_nodes_count}")
            if valid_node_count == self.cluster_nodes_count:
                self.cluster_status = ClusterStatus.READY
                logger.info(f"cluster of {self.cluster_nodes_count} nodes is ready")
        elif self.cluster_status == ClusterStatus.READY:
            if valid_node_count < self.cluster_nodes_count:
                self.cluster_status = ClusterStatus.TEARING_DOWN
                logger.info(f"cluster of {self.cluster_nodes_count} nodes is being torn down")
        elif self.cluster_status == ClusterStatus.TEARING_DOWN:
            if valid_node_count == 0:
                self.cluster_status = ClusterStatus.FINISHED
    
    def is_cluster_ready(self):
        return self.cluster_status == ClusterStatus.READY
    
    def is_cluster_tearing_down(self):
        return self.cluster_status == ClusterStatus.TEARING_DOWN
    
    def is_cluster_finished(self):
        valid_node_count = self.count_valid_nodes()
        if valid_node_count == 0:
            self.cluster_status = ClusterStatus.FINISHED
        return self.cluster_status == ClusterStatus.FINISHED







import os
import ray
# import socket
import random
from loguru import logger

class BaseBackend:
    def __init__(self):
        self.backend_name = self.__class__.__name__ + str(id(self))

    def get_gpu_ids(self):
        return ray.get_gpu_ids()
    
    def get_ip_address(self):
        return ray.util.get_node_ip_address()
    
    def get_avaiable_port(self):
        port = random.randint(58000, 62000)
        return port
        # while True:
        #     port = random.randint(58000, 62000)
        #     with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        #         is_valid = s.connect_ex(('localhost', port)) == 0
        #     if is_valid:
        #         return port


    def set_environ(self, name, val):
        old_val =  os.environ.get(name)
        if old_val:
            logger.debug(f"environ {name} of backend {self.backend_name} is already a non-empty value ({old_val}), now overriding with {val}")
        os.environ[name] = str(val)
    
    def set_distributed_environs(
            self,
            rank,
            world_size,
            master_addr,
            master_port
    ):
        self.set_environ("RANK", rank)
        gpu_ids = self.get_gpu_ids()
        assert len(gpu_ids) == 1
        gpu_id = gpu_ids.pop()
        self.set_environ("LOCAL_RANK", gpu_id)
        self.set_environ("WORLD_SIZE", world_size)
        self.set_environ("MASTER_ADDR", master_addr)
        self.set_environ("MASTER_PORT", master_port)

    def format_distributed_info(self):
        s = "distributed environs: "
        for e_name in ["RANK", "WORLD_SIZE", "MASTER_ADDR", "MASTER_PORT", "LOCAL_RANK"]:
            s += f"{e_name}={os.environ.get(e_name)} "
        return s



import os
import time
from ray_launcher import ClusterLauncher, BaseBackend, RemoteModule
import ray

os.environ["RAY_DISABLE_DOCKER_CPU_WARNING"] = "1"
os.environ["RAY_memory_monitor_refresh_ms"] = "0"
os.environ["RAY_DEDUP_LOGS"] = "0"

class MockBackend(BaseBackend):
    def get_devices(self):
        return self.backend_name + ": " + os.environ.get("CUDA_VISIBLE_DEVICES")
        

with ClusterLauncher(
    cluster_nodes_count=int(os.environ["NNODES"]),
    head_node_addr=os.environ["MASTER_ADDR"],
    export_env_var_names=["RAY_DISABLE_DOCKER_CPU_WARNING"]
) as launcher:
    print("cluster ready")
    assert launcher.is_head_node, f"only head node reaches here"
    bundle = [{"GPU": 2, "CPU": 32}, {"GPU": 2, "CPU": 32}]
    pg = ray.util.placement_group(bundle, strategy="PACK")
    module1 = RemoteModule(MockBackend, [(pg, 0)], True)
    module2 = RemoteModule(MockBackend, [(pg, 1)], False)

    print(module1.get_devices())
    print(module2.get_devices())

    time.sleep(5)
    print("prepare to stop the cluster")